from asyncio import iscoroutinefunction
from datetime import datetime, timedelta, timezone
from functools import wraps
from inspect import isgeneratorfunction, isasyncgenfunction, isclass
from math import ceil, floor
from time import monotonic
from typing import AnyStr, Iterable

STRING_TYPES = (bytes, str)
STATE_CLOSED = 'closed'
STATE_OPEN = 'open'
STATE_HALF_OPEN = 'half_open'


def in_exception_list(*exc_types):
    """Build a predicate function that checks if an exception is a subtype from a list"""

    def matches_types(thrown_type, _):
        return issubclass(thrown_type, exc_types)

    return matches_types


def build_failure_predicate(expected_exception):
    """ Build a failure predicate_function.
          The returned function has the signature (Type[Exception], Exception) -> bool.
          Return value True indicates a failure in the underlying function.

        :param expected_exception: either an type of Exception, iterable of Exception types, or a predicate function.

          If an Exception type or iterable of Exception types, the failure predicate will return True when a thrown
          exception type matches one of the provided types.

          If a predicate function, it will just be returned as is.

         :return: callable (Type[Exception], Exception) -> bool
    """

    if isclass(expected_exception) and issubclass(expected_exception, Exception):
        failure_predicate = in_exception_list(expected_exception)
    else:
        try:
            # Check for an iterable of Exception types
            iter(expected_exception)

            # guard against a surprise later
            if isinstance(expected_exception, STRING_TYPES):
                raise ValueError("expected_exception cannot be a string. Did you mean name?")
            failure_predicate = in_exception_list(*expected_exception)
        except TypeError:
            # not iterable. guess that it's a predicate function
            if not callable(expected_exception) or isclass(expected_exception):
                raise ValueError("expected_exception does not look like a predicate")
            failure_predicate = expected_exception
    return failure_predicate


class CircuitBreaker(object):
    FAILURE_THRESHOLD = 5
    RECOVERY_TIMEOUT = 30
    EXPECTED_EXCEPTION = Exception
    FALLBACK_FUNCTION = None

    def __init__(self,
                 failure_threshold=None,
                 recovery_timeout=None,
                 expected_exception=None,
                 name=None,
                 fallback_function=None
                 ):
        """
        Construct a circuit breaker.

        :param failure_threshold: break open after this many failures
        :param recovery_timeout: close after this many seconds
        :param expected_exception: either an type of Exception, iterable of Exception types, or a predicate function.
        :param name: name for this circuitbreaker
        :param fallback_function: called when the circuit is opened

           :return: Circuitbreaker instance
           :rtype: Circuitbreaker
        """
        self._last_failure = None
        self._failure_count = 0
        self._failure_threshold = failure_threshold or self.FAILURE_THRESHOLD
        self._recovery_timeout = recovery_timeout or self.RECOVERY_TIMEOUT

        # Build the failure predicate. In order of precedence, prefer the
        # * the constructor argument
        # * the subclass attribute EXPECTED_EXCEPTION
        # * the CircuitBreaker attribute EXPECTED_EXCEPTION
        if not expected_exception:
            try:
                # Introspect our final type, then grab the  value via __dict__ to avoid python Descriptor magic
                #  in the case where it's a callable function.
                expected_exception = type(self).__dict__["EXPECTED_EXCEPTION"]
            except KeyError:
                expected_exception = CircuitBreaker.EXPECTED_EXCEPTION

        self.is_failure = build_failure_predicate(expected_exception)

        self._fallback_function = fallback_function or self.FALLBACK_FUNCTION
        self._name = name
        self._state = STATE_CLOSED
        self._opened = monotonic()

    def __call__(self, wrapped):
        return self.decorate(wrapped)

    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_value, _traceback):
        if exc_type and self.is_failure(exc_type, exc_value):
            # exception was raised and is our concern
            self._last_failure = exc_value
            self.__call_failed()
        else:
            self.reset()
        return False  # return False to raise exception if any

    def decorate(self, function):
        """
        Applies the circuit breaker to a function
        """
        if self._name is None:
            try:
                self._name = function.__qualname__
            except AttributeError:
                self._name = function.__name__

        CircuitBreakerMonitor.register(self)

        if iscoroutinefunction(function) or isasyncgenfunction(function):
            return self._decorate_async(function)

        return self._decorate_sync(function)

    def _decorate_sync(self, function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            if self.opened:
                if self.fallback_function:
                    return self.fallback_function(*args, **kwargs)
                raise CircuitBreakerError(self)
            return self.call(function, *args, **kwargs)

        @wraps(function)
        def gen_wrapper(*args, **kwargs):
            if self.opened:
                if self.fallback_function:
                    yield from self.fallback_function(*args, **kwargs)
                    return
                raise CircuitBreakerError(self)
            yield from self.call_generator(function, *args, **kwargs)

        return gen_wrapper if isgeneratorfunction(function) else wrapper

    def _decorate_async(self, function):
        @wraps(function)
        async def awrapper(*args, **kwargs):
            if self.opened:
                if self.fallback_function:
                    return await self.fallback_function(*args, **kwargs)
                raise CircuitBreakerError(self)
            return await self.call_async(function, *args, **kwargs)

        @wraps(function)
        async def gen_awrapper(*args, **kwargs):
            if self.opened:
                if self.fallback_function:
                    async for el in self.fallback_function(*args, **kwargs):
                        yield el
                    return
                raise CircuitBreakerError(self)
            async for el in self.call_async_generator(function, *args, **kwargs):
                yield el

        return gen_awrapper if isasyncgenfunction(function) else awrapper

    def call(self, func, *args, **kwargs):
        """
        Calls the decorated function and applies the circuit breaker
        rules on success or failure
        :param func: Decorated function
        """
        with self:
            return func(*args, **kwargs)

    def call_generator(self, func, *args, **kwargs):
        """
        Calls the decorated generator function and applies the circuit breaker
        rules on success or failure
        :param func: Decorated generator function
        """
        with self:
            for el in func(*args, **kwargs):
                yield el

    async def call_async(self, func, *args, **kwargs):
        """
        Calls the decorated async function and applies the circuit breaker
        rules on success or failure
        :param func: Decorated async function
        """
        with self:
            return await func(*args, **kwargs)

    async def call_async_generator(self, func, *args, **kwargs):
        """
        Calls the decorated async generator function and applies the circuit breaker
        rules on success or failure
        :param func: Decorated async generator function
        """
        with self:
            async for el in func(*args, **kwargs):
                yield el

    def reset(self):
        """
        Close circuit and reset failure count
        """
        self._state = STATE_CLOSED
        self._last_failure = None
        self._failure_count = 0

    def __call_failed(self):
        """
        Count failure and open circuit, if threshold has been reached
        """
        self._failure_count += 1
        if self._failure_count >= self._failure_threshold:
            self._state = STATE_OPEN
            self._opened = monotonic()

    @property
    def state(self):
        if self._state == STATE_OPEN and self.open_remaining <= 0:
            return STATE_HALF_OPEN
        return self._state

    @property
    def open_until(self):
        """
        The approximate datetime when the circuit breaker will try to recover
        :return: datetime
        """
        return datetime.now(timezone.utc) + timedelta(seconds=self.open_remaining)

    @property
    def open_remaining(self):
        """
        Number of seconds remaining, the circuit breaker stays in OPEN state
        :return: int
        """
        remain = (self._opened + self._recovery_timeout) - monotonic()
        return ceil(remain) if remain > 0 else floor(remain)

    @property
    def failure_count(self):
        return self._failure_count

    @property
    def closed(self):
        return self.state == STATE_CLOSED

    @property
    def opened(self):
        return self.state == STATE_OPEN

    @property
    def name(self):
        return self._name

    @property
    def last_failure(self):
        return self._last_failure

    @property
    def fallback_function(self):
        return self._fallback_function

    def __str__(self, *args, **kwargs):
        return self._name


class CircuitBreakerError(Exception):
    def __init__(self, circuit_breaker, *args, **kwargs):
        """
        :param circuit_breaker:
        :param args:
        :param kwargs:
        :return:
        """
        super(CircuitBreakerError, self).__init__(*args, **kwargs)
        self._circuit_breaker = circuit_breaker

    def __str__(self, *args, **kwargs):
        return 'Circuit "%s" OPEN until %s (%d failures, %d sec remaining) (last_failure: %r)' % (
            self._circuit_breaker.name,
            self._circuit_breaker.open_until,
            self._circuit_breaker.failure_count,
            round(self._circuit_breaker.open_remaining),
            self._circuit_breaker.last_failure,
        )


class CircuitBreakerMonitor(object):
    circuit_breakers = {}

    @classmethod
    def register(cls, circuit_breaker):
        cls.circuit_breakers[circuit_breaker.name] = circuit_breaker

    @classmethod
    def all_closed(cls) -> bool:
        return len(list(cls.get_open())) == 0

    @classmethod
    def get_circuits(cls) -> Iterable[CircuitBreaker]:
        return cls.circuit_breakers.values()

    @classmethod
    def get(cls, name: AnyStr) -> CircuitBreaker:
        return cls.circuit_breakers.get(name)

    @classmethod
    def get_open(cls) -> Iterable[CircuitBreaker]:
        for circuit in cls.get_circuits():
            if circuit.opened:
                yield circuit

    @classmethod
    def get_closed(cls) -> Iterable[CircuitBreaker]:
        for circuit in cls.get_circuits():
            if circuit.closed:
                yield circuit


def circuit(failure_threshold=None,
            recovery_timeout=None,
            expected_exception=None,
            name=None,
            fallback_function=None,
            cls=CircuitBreaker):
    # if the decorator is used without parameters, the
    # wrapped function is provided as first argument
    if callable(failure_threshold):
        return cls().decorate(failure_threshold)
    else:
        return cls(
            failure_threshold=failure_threshold,
            recovery_timeout=recovery_timeout,
            expected_exception=expected_exception,
            name=name,
            fallback_function=fallback_function)



import asyncio
import time
from enum import Enum

import pytest

from circuitbreaker import CircuitBreakerMonitor


class FunctionType(str, Enum):
    sync_function = "sync-function"
    sync_generator = "sync-generator"
    async_function = "async-function"
    async_generator = "async-generator"


@pytest.fixture(autouse=True)
def clean_circuit_breaker_monitor():
    CircuitBreakerMonitor.circuit_breakers = {}


@pytest.fixture(params=FunctionType, ids=[e.value for e in FunctionType])
def function_type(request):
    return request.param


@pytest.fixture
def is_async(function_type):
    return function_type.startswith("async-")


@pytest.fixture
def is_generator(function_type):
    return function_type.endswith("-generator")


@pytest.fixture
def function_factory(function_type):
    def factory(inner_call):
        def _sync(*a, **kwa):
            return inner_call(*a, **kwa)

        def _sync_gen(*a, **kwa):
            yield inner_call(*a, **kwa)

        async def _async(*a, **kwa):
            return inner_call(*a, **kwa)

        async def _async_gen(*a, **kwa):
            yield inner_call(*a, **kwa)

        mapping = {
            FunctionType.sync_function: _sync,
            FunctionType.sync_generator: _sync_gen,
            FunctionType.async_function: _async,
            FunctionType.async_generator: _async_gen,
        }
        return mapping[function_type]

    return factory


@pytest.fixture
def resolve_call(function_type):
    """
    This fixture helps abstract calls from other fixtures that have sync and
    async, function and generator versions.

    For example, this:
        if is_async:
            if is_generator:
                result = [el async for el in function()]
            else:
                result = await function()
        else:
            if is_generator:
                result = list(function())
            else:
                result = function()

    Can be replaced with:
        result = await resolve_call(function())

    """
    async def _sync(value):
        return value

    async def _sync_gen(generator):
        return list(generator)

    async def _async(coroutine):
        return await coroutine

    async def _async_gen(async_generator):
        return [el async for el in async_generator]

    mapping = {
        FunctionType.sync_function: _sync,
        FunctionType.sync_generator: _sync_gen,
        FunctionType.async_function: _async,
        FunctionType.async_generator: _async_gen,
    }
    return mapping[function_type]


@pytest.fixture
def sleep(is_async):
    async def _sleep(secs):
        if is_async:
            await asyncio.sleep(secs)
        else:
            time.sleep(secs)

    return _sleep


@pytest.fixture
def mock_function_call(mocker):
    return mocker.Mock(return_value=object())


@pytest.fixture
def mock_fallback_call(mocker):
    return mocker.Mock(return_value=object())


@pytest.fixture
def function_call_return_value(is_generator, mock_function_call):
    value = mock_function_call.return_value
    return [value] if is_generator else value


@pytest.fixture
def fallback_call_return_value(is_generator, mock_fallback_call):
    value = mock_fallback_call.return_value
    return [value] if is_generator else value


@pytest.fixture
def function_call_error(mock_function_call):
    error = IOError
    mock_function_call.side_effect = error
    return error


@pytest.fixture
def function(function_factory, mock_function_call):
    return function_factory(mock_function_call)


@pytest.fixture
def fallback_function(function_factory, mock_fallback_call):
    return function_factory(mock_fallback_call)



import pytest

from circuitbreaker import (
    CircuitBreaker,
    CircuitBreakerError,
    CircuitBreakerMonitor,
    STATE_CLOSED,
    STATE_HALF_OPEN,
    STATE_OPEN,
)


@pytest.fixture
def circuit_success(function):
    return CircuitBreaker()(function)


@pytest.fixture
def circuit_failure(function, function_call_error):
    return CircuitBreaker(
        failure_threshold=1,
        name="circuit_failure",
    )(function)


@pytest.fixture
def circuit_threshold_1(function):
    return CircuitBreaker(
        failure_threshold=1,
        name="threshold_1",
    )(function)


@pytest.fixture
def circuit_threshold_2_timeout_1(function):
    return CircuitBreaker(
        failure_threshold=2,
        recovery_timeout=1,
        name="threshold_2",
    )(function)


@pytest.fixture
def circuit_threshold_3_timeout_1(function):
    return CircuitBreaker(
        failure_threshold=3,
        recovery_timeout=1,
        name="threshold_3",
    )(function)


async def test_circuit_pass_through(
    resolve_call, circuit_success, function_call_return_value
):
    assert await resolve_call(circuit_success()) == function_call_return_value


@pytest.mark.usefixtures(
    "circuit_success",
    "circuit_failure",
    "circuit_threshold_1",
    "circuit_threshold_2_timeout_1",
    "circuit_threshold_3_timeout_1",
)
async def test_circuitbreaker_monitor(
    resolve_call, circuit_failure, function_call_error
):
    assert CircuitBreakerMonitor.all_closed() is True
    assert len(list(CircuitBreakerMonitor.get_circuits())) == 5
    assert len(list(CircuitBreakerMonitor.get_closed())) == 5
    assert len(list(CircuitBreakerMonitor.get_open())) == 0

    with pytest.raises(function_call_error):
        await resolve_call(circuit_failure())

    assert CircuitBreakerMonitor.all_closed() is False
    assert len(list(CircuitBreakerMonitor.get_circuits())) == 5
    assert len(list(CircuitBreakerMonitor.get_closed())) == 4
    assert len(list(CircuitBreakerMonitor.get_open())) == 1


async def test_threshold_hit_prevents_consequent_calls(
    resolve_call, mock_function_call, circuit_threshold_1, function_call_error
):
    circuitbreaker = CircuitBreakerMonitor.get('threshold_1')

    assert circuitbreaker.closed

    with pytest.raises(function_call_error):
        await resolve_call(circuit_threshold_1())

    assert circuitbreaker.opened

    with pytest.raises(CircuitBreakerError):
        await resolve_call(circuit_threshold_1())

    mock_function_call.assert_called_once_with()


async def test_circuitbreaker_recover_half_open(
    resolve_call, mock_function_call, circuit_threshold_3_timeout_1, sleep
):
    circuitbreaker = CircuitBreakerMonitor.get('threshold_3')

    # initial state: closed
    assert circuitbreaker.closed
    assert circuitbreaker.state == STATE_CLOSED

    # no exception -> success
    assert await resolve_call(circuit_threshold_3_timeout_1())

    # from now all subsequent calls will fail
    mock_function_call.side_effect = IOError('Connection refused')

    # 1. failed call -> original exception
    with pytest.raises(IOError):
        assert await resolve_call(circuit_threshold_3_timeout_1())

    assert circuitbreaker.closed
    assert circuitbreaker.failure_count == 1

    # 2. failed call -> original exception
    with pytest.raises(IOError):
        assert await resolve_call(circuit_threshold_3_timeout_1())
    assert circuitbreaker.closed
    assert circuitbreaker.failure_count == 2

    # 3. failed call -> original exception
    with pytest.raises(IOError):
        assert await resolve_call(circuit_threshold_3_timeout_1())

    # Circuit breaker opens, threshold has been reached
    assert circuitbreaker.opened
    assert circuitbreaker.state == STATE_OPEN
    assert circuitbreaker.failure_count == 3
    assert 0 < circuitbreaker.open_remaining <= 1

    # 4. failed call -> not passed to function -> CircuitBreakerError
    with pytest.raises(CircuitBreakerError):
        assert await resolve_call(circuit_threshold_3_timeout_1())
    assert circuitbreaker.opened
    assert circuitbreaker.failure_count == 3
    assert 0 < circuitbreaker.open_remaining <= 1

    # 5. failed call -> not passed to function -> CircuitBreakerError
    with pytest.raises(CircuitBreakerError):
        assert await resolve_call(circuit_threshold_3_timeout_1())
    assert circuitbreaker.opened
    assert circuitbreaker.failure_count == 3
    assert 0 < circuitbreaker.open_remaining <= 1

    # wait for 1 second (recover timeout)
    await sleep(1)

    # circuit half-open -> next call will be passed through
    assert not circuitbreaker.closed
    assert circuitbreaker.open_remaining < 0
    assert circuitbreaker.state == STATE_HALF_OPEN

    # State half-open -> function is executed -> original exception
    with pytest.raises(IOError):
        assert await resolve_call(circuit_threshold_3_timeout_1())
    assert circuitbreaker.opened
    assert circuitbreaker.failure_count == 4
    assert 0 < circuitbreaker.open_remaining <= 1

    # State open > not passed to function -> CircuitBreakerError
    with pytest.raises(CircuitBreakerError):
        assert await resolve_call(circuit_threshold_3_timeout_1())


async def test_circuitbreaker_reopens_after_successful_calls(
    resolve_call, mock_function_call, circuit_threshold_2_timeout_1, sleep
):
    circuitbreaker = CircuitBreakerMonitor.get('threshold_2')

    assert str(circuitbreaker) == 'threshold_2'

    # initial state: closed
    assert circuitbreaker.closed
    assert circuitbreaker.state == STATE_CLOSED
    assert circuitbreaker.failure_count == 0

    # successful call -> no exception
    assert await resolve_call(circuit_threshold_2_timeout_1())

    # from now all subsequent calls will fail
    mock_function_call.side_effect = IOError('Connection refused')

    # 1. failed call -> original exception
    with pytest.raises(IOError):
        await resolve_call(circuit_threshold_2_timeout_1())
    assert circuitbreaker.closed
    assert circuitbreaker.failure_count == 1

    # 2. failed call -> original exception
    with pytest.raises(IOError):
        await resolve_call(circuit_threshold_2_timeout_1())

    # Circuit breaker opens, threshold has been reached
    assert circuitbreaker.opened
    assert circuitbreaker.state == STATE_OPEN
    assert circuitbreaker.failure_count == 2
    assert 0 < circuitbreaker.open_remaining <= 1

    # 4. failed call -> not passed to function -> CircuitBreakerError
    with pytest.raises(CircuitBreakerError):
        await resolve_call(circuit_threshold_2_timeout_1())
    assert circuitbreaker.opened
    assert circuitbreaker.failure_count == 2
    assert 0 < circuitbreaker.open_remaining <= 1

    # from now all subsequent calls will succeed
    mock_function_call.side_effect = None

    # but recover timeout has not been reached -> still open
    # 5. failed call -> not passed to function -> CircuitBreakerError
    with pytest.raises(CircuitBreakerError):
        await resolve_call(circuit_threshold_2_timeout_1())
    assert circuitbreaker.opened
    assert circuitbreaker.failure_count == 2
    assert 0 < circuitbreaker.open_remaining <= 1

    # wait for 1 second (recover timeout)
    await sleep(1)

    # circuit half-open -> next call will be passed through
    assert not circuitbreaker.closed
    assert circuitbreaker.failure_count == 2
    assert circuitbreaker.open_remaining < 0
    assert circuitbreaker.state == STATE_HALF_OPEN

    # successful call
    assert await resolve_call(circuit_threshold_2_timeout_1())

    # circuit closed and reset'ed
    assert circuitbreaker.closed
    assert circuitbreaker.state == STATE_CLOSED
    assert circuitbreaker.failure_count == 0

    # some another successful calls
    assert await resolve_call(circuit_threshold_2_timeout_1())
    assert await resolve_call(circuit_threshold_2_timeout_1())
    assert await resolve_call(circuit_threshold_2_timeout_1())



from asyncio import iscoroutinefunction
from inspect import isgeneratorfunction, isasyncgenfunction

import pytest

from circuitbreaker import CircuitBreaker, CircuitBreakerError, circuit


@pytest.fixture
def resolve_circuitbreaker_call_method(function_type):
    def cb_call(circuit_breaker):
        mapping = {
            "sync-function": circuit_breaker.call,
            "sync-generator": circuit_breaker.call_generator,
            "async-function": circuit_breaker.call_async,
            "async-generator": circuit_breaker.call_async_generator,
        }
        return mapping[function_type]

    return cb_call


class FooError(Exception):
    def __init__(self, val=None):
        self.val = val


class BarError(Exception):
    pass


def test_circuitbreaker__str__():
    cb = CircuitBreaker(name='Foobar')
    assert str(cb) == 'Foobar'


def test_circuitbreaker_error__str__():
    cb = CircuitBreaker(name='Foobar')
    cb._last_failure = Exception()
    error = CircuitBreakerError(cb)

    assert str(error).startswith('Circuit "Foobar" OPEN until ')
    assert str(error).endswith('(0 failures, 30 sec remaining) (last_failure: Exception())')


async def test_circuitbreaker_wrapper_matches_function_type(function, is_async, is_generator):
    cb = CircuitBreaker(name='Foobar')
    wrapper = cb(function)

    assert (
        isgeneratorfunction(function) == isgeneratorfunction(wrapper)
        == (not is_async and is_generator)
    )
    assert (
        iscoroutinefunction(function) == iscoroutinefunction(wrapper)
        == (is_async and not is_generator)
    )
    assert (
        isasyncgenfunction(function) == isasyncgenfunction(wrapper)
        == (is_async and is_generator)
    )


async def test_circuitbreaker_should_save_last_exception_on_failure_call(
    resolve_call, resolve_circuitbreaker_call_method, function, function_call_error
):
    cb = CircuitBreaker(name='Foobar')

    cb_call = resolve_circuitbreaker_call_method(cb)
    with pytest.raises(function_call_error):
        await resolve_call(cb_call(function))

    assert isinstance(cb.last_failure, function_call_error)


async def test_circuitbreaker_should_clear_last_exception_on_success_call(
    resolve_call, resolve_circuitbreaker_call_method, function
):
    cb = CircuitBreaker(name='Foobar')
    cb._last_failure = IOError()
    assert isinstance(cb.last_failure, IOError)

    cb_call = resolve_circuitbreaker_call_method(cb)
    await resolve_call(cb_call(function))

    assert cb.last_failure is None


async def test_circuitbreaker_should_call_fallback_function_if_open(
    resolve_call, function, fallback_function, mock_fallback_call, fallback_call_return_value
):
    CircuitBreaker.opened = lambda self: True

    cb = CircuitBreaker(name='WithFallback', fallback_function=fallback_function)
    decorated_func = cb.decorate(function)

    assert await resolve_call(decorated_func()) == fallback_call_return_value
    mock_fallback_call.assert_called_once_with()


async def test_circuitbreaker_should_not_call_function_if_open(
    resolve_call, function, mock_function_call, fallback_function, fallback_call_return_value
):
    CircuitBreaker.opened = lambda self: True

    cb = CircuitBreaker(name='WithFallback', fallback_function=fallback_function)
    decorated_func = cb.decorate(function)

    assert await resolve_call(decorated_func()) == fallback_call_return_value
    assert not mock_function_call.called


async def test_circuitbreaker_call_fallback_function_with_parameters(
    resolve_call, function, fallback_function, mock_fallback_call, fallback_call_return_value
):
    # mock opened prop to see if fallback is called with correct parameters.
    CircuitBreaker.opened = lambda self: True

    cb = CircuitBreaker(name='WithFallback', fallback_function=fallback_function)
    decorated_func = cb.decorate(function)

    assert await resolve_call(decorated_func('test2', test='test')) == fallback_call_return_value

    # check args and kwargs are getting correctly to fallback function
    mock_fallback_call.assert_called_once_with('test2', test='test')


def test_circuit_decorator_without_args(mocker, function):
    decorate_patch = mocker.patch.object(CircuitBreaker, 'decorate')
    circuit(function)
    decorate_patch.assert_called_once_with(function)


def test_circuit_decorator_with_args():
    def function_fallback():
        return True

    breaker = circuit(10, 20, KeyError, 'foobar', function_fallback)

    assert breaker.is_failure(KeyError, None)
    assert not breaker.is_failure(Exception, None)
    assert not breaker.is_failure(FooError, None)
    assert breaker._failure_threshold == 10
    assert breaker._recovery_timeout == 20
    assert breaker._name == "foobar"
    assert breaker._fallback_function == function_fallback


def test_breaker_expected_exception_is_predicate():
    def is_four_foo(thrown_type, thrown_value):
        return thrown_value.val == 4

    breaker_four = circuit(expected_exception=is_four_foo)

    assert breaker_four.is_failure(FooError, FooError(4))
    assert not breaker_four.is_failure(FooError, FooError(2))


def test_breaker_default_constructor_traps_Exception():
    breaker = circuit()
    assert breaker.is_failure(Exception, Exception())
    assert breaker.is_failure(FooError, FooError())


def test_breaker_expected_exception_is_custom_exception():
    breaker = circuit(expected_exception=FooError)
    assert breaker.is_failure(FooError, FooError())
    assert not breaker.is_failure(Exception, Exception())


def test_breaker_constructor_expected_exception_is_exception_list():
    breaker = circuit(expected_exception=(FooError, BarError))
    assert breaker.is_failure(FooError, FooError())
    assert breaker.is_failure(BarError, BarError())
    assert not breaker.is_failure(Exception, Exception())


def test_constructor_mistake_name_bytes():
    with pytest.raises(ValueError, match="expected_exception cannot be a string *"):
        circuit(10, 20, b"foobar")


def test_constructor_mistake_name_unicode():
    with pytest.raises(ValueError, match="expected_exception cannot be a string *"):
        circuit(10, 20, u"foobar")


def test_constructor_mistake_expected_exception():
    class Widget:
        pass

    with pytest.raises(ValueError, match="expected_exception does not look like a predicate*"):
        circuit(10, 20, expected_exception=Widget)


def test_advanced_usage_circuitbreaker_subclass():
    class MyCircuitBreaker(CircuitBreaker):
        EXPECTED_EXCEPTION = FooError

    mybreaker = circuit(cls=MyCircuitBreaker)
    assert not mybreaker.is_failure(Exception, Exception())
    assert mybreaker.is_failure(FooError, FooError())


def test_advanced_usage_circuitbreaker_subclass_with_list():
    class MyCircuitBreaker(CircuitBreaker):
        EXPECTED_EXCEPTION = (FooError, BarError)

    mybreaker = circuit(cls=MyCircuitBreaker)
    assert not mybreaker.is_failure(Exception, Exception())
    assert mybreaker.is_failure(FooError, FooError())
    assert mybreaker.is_failure(BarError, BarError())


def test_advanced_usage_circuitbreaker_subclass_with_predicate():
    def is_foo_4(thrown_type, thrown_value):
        return issubclass(thrown_type, FooError) and thrown_value.val == 4

    class FooFourBreaker(CircuitBreaker):
        EXPECTED_EXCEPTION = is_foo_4

    breaker = circuit(cls=FooFourBreaker)
    assert not breaker.is_failure(Exception, Exception())
    assert not breaker.is_failure(FooError, FooError())
    assert breaker.is_failure(FooError, FooError(4))


def test_advanced_usage_circuitbreaker_default_expected_exception():
    class NervousBreaker(CircuitBreaker):
        FAILURE_THRESHOLD = 1

    breaker = circuit(cls=NervousBreaker)
    assert breaker._failure_threshold == 1
    assert breaker.is_failure(Exception, Exception())
